/**
  *  \file client/dlg-build.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-build.h"
#include "client/winfo.h"
#include "game/planet.h"
#include "game/spec.h"
#include "game/struct.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/table.h"
#include "ui/widgets.h"
#include "util/misc.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/msgbox.h"
#include "resmgr/resid.h"
#include "ui/util.h"
#include "client/dlg-component.h"
#include "game/global.h"
#include "client/wcomplist.h"
#include "game/shipquery.h"
#include "client/widgets/specview.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/statictext.h"
#include "resmgr/resmgr.h"

enum {
    cm_AddComponent    = 4000,
    cm_RemoveComponent = 4001,
    cm_CountChanged    = 4002,
    cm_BuildComponent  = 4003
};

/***************************** WOrderDisplay *****************************/

/** \class WOrderDisplay
    \brief Display a Build Order

    This will display a build order on the ship build screen. */
WOrderDisplay::WOrderDisplay()
{
    o.hull = 0;
}

void
WOrderDisplay::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getExtent(), FONT_NORMAL);

    if (o.hull) {
        GHull& h = getHull(o.hull);

        tw.newLColumn();
        tw.writeVariableCell(UIColor::tc_Contrast, _("Your order:"));
        tw.writeVariableCell(UIColor::tc_Static, "\xE2\x80\xA2 " + h.getName());
        tw.writeVariableCell(UIColor::tc_Static,
                             /* this format string turns into
                                "Transwarp Drive" or "7 x Transwarp Drive" */
                             format(_("\xE2\x80\xA2 %!d%!1{%0$d \xC3\x97 %}%1$s"),
                                    h.getNumEngines(),
                                    getEngine(o.engine).getName()));
        if (o.beam_count)
            tw.writeVariableCell(UIColor::tc_Static,
                                 format(_("\xE2\x80\xA2 %!d%!1{%0$d \xC3\x97 %}%1$s"),
                                        o.beam_count,
                                        getBeam(o.beam_type).getName()));
        if (o.torp_count)
            tw.writeVariableCell(UIColor::tc_Static,
                                 format(_("\xE2\x80\xA2 %!d%!1{%0$d \xC3\x97 %}%1$s"),
                                        o.torp_count,
                                        getTorp(o.torp_type).getName()));
        if (h.getNumBays())
            tw.writeVariableCell(UIColor::tc_Static,
                                 format(_("\xE2\x80\xA2 %d fighter bay%!1{s%}"), h.getNumBays()));
    }
    tw.finish();
    can.update(getExtent());
}

void
WOrderDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(20, 5);
}

void
WOrderDisplay::setBuildOrder(const TBuildOrder& o)
{
    this->o = o;
    drawWidget();
}

static UIBaseWidget&
addFrame(ObjectHolder& h, UIBaseWidget& w)
{
    UIFrameGroup& fg = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    fg.add(w);
    return fg;
}

/************************** WPartStorageDisplay **************************/

WPartStorageDisplay::WPartStorageDisplay()
    : UIWidget(), n(0)
{ }

void
WPartStorageDisplay::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_Static);
    outTextF(ctx, getExtent().x, getExtent().y, getExtent().w,
             _("In storage:"));
    outTextF(ctx, getExtent().x, getExtent().y + font_heights[FONT_NORMAL], getExtent().w,
             format(_("%d pcs."), n));
    can.update(getExtent());
}

void
WPartStorageDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(7, 2);
    info.flags = 0;
}

void
WPartStorageDisplay::setCount(int n)
{
    if (n != this->n) {
        this->n = n;
        drawWidget(false);
    }
}


WPartStorageGroup::WPartStorageGroup()
    : UIGroup(UIVBoxLayout::instance5),
      image(105+2, 93+2 /* FIXME! Make this variable? */),
      psd(),
      button(cm_BuildComponent, _("Spc-Build"), ' ')
{
    add(image);
    add(psd);
    // add(spacer);   FIXME: do we need it?
    add(button);
    button.addCommand();
}

WPartStorageGroup::~WPartStorageGroup()
{ }

/***************************** WPartCountInfo ****************************/

WPartCountInfo::WPartCountInfo(bool torps)
    : torps(torps), want(0), max(0)
{ }

void
WPartCountInfo::drawContent(GfxCanvas& can)
{
    const char* fmt = torps ? _("Max. launchers on ship: %d") : _("Max. beams on ship: %d");
    int x = getExtent().x, y = getExtent().y, wi = getExtent().w;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_White);

    outTextF(ctx, x, y, wi, format(fmt, max));
    y += font_heights[FONT_NORMAL];
    if (max > 0) {
        outTextF(ctx, x, y, wi, format(_("Actual number: %d"), want));
    } else {
        outTextF(ctx, x, y, wi, string_t());
    }
    can.update(getExtent());
}

void
WPartCountInfo::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(15, 2);
    info.flags = li_FlexibleH;
}

void
WPartCountInfo::setPartsMax(int n)
{
    if (n != max) {
        max = n;
        want = max;
        drawWidget(false);
    }
}

void
WPartCountInfo::setPartsWanted(int n)
{
    if (n != want) {
        want = n;
        drawWidget(false);
    }
}

WPartCountGroup::WPartCountGroup(bool torps)
    : UIGroup(UIHBoxLayout::instance0),
      pci(torps),
      button_group(UIVBoxLayout::instance0),
      plus(cm_AddComponent, "+", '+'),
      minus(cm_RemoveComponent, "-", '-')
{
    add(pci);
    add(button_group);

    button_group.add(plus);
    button_group.add(minus);

    plus.addCommand();
    plus.setState(UIButton::bf_SmallSize, true);
    minus.addCommand();
    minus.setState(UIButton::bf_SmallSize, true);
}

WPartCountGroup::~WPartCountGroup()
{ }

void
WPartCountGroup::setPartsMax(int n)
{
    pci.setPartsMax(n);
    setStateOf(cm_AddComponent, st_Disabled, n == 0);
    setStateOf(cm_RemoveComponent, st_Disabled, n == 0);
}

bool
WPartCountGroup::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_AddComponent) {
        int n = getPartsWanted() + (cmd.arg ? cmd.arg : 1);
        if (n > getPartsMax())
            n = getPartsMax();
        setPartsWanted(n);
        handleCommand(UICommand(cm_CountChanged), *this);
        return true;
    } else if (cmd.code == cm_RemoveComponent) {
        int n = getPartsWanted() - (cmd.arg ? cmd.arg : 1);
        if (n < 0)
            n = 0;
        setPartsWanted(n);
        handleCommand(UICommand(cm_CountChanged), *this);
        return true;
    } else {
        return UIGroup::handleCommand(cmd, sender);
    }
}

/*************************** WEngineInfoDiagram **************************/

WEngineInfoDiagram::WEngineInfoDiagram()
    : UIWidget(0), engine(1)
{ }

void
WEngineInfoDiagram::drawContent(GfxCanvas& can)
{
    int bottom = getExtent().getBottomY();
    int center = getExtent().getCenter().x;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_SMALL]);
    ctx.setTextAlign(1, 0);
    ctx.setColor(UIColor::tc_Static);

    string_t desc = _("Fuel used per\nlightyear");

    outTextFormatted(ctx, center, bottom - 2*font_heights[FONT_SMALL], extent.w, desc);

    GfxRect diagram_pos(getExtent());
    diagram_pos.h -= 2*font_heights[FONT_SMALL] + 1;

    drawFrameDown(can, diagram_pos);
    diagram_pos.grow(-1, -1);
    drawSolidBar(can, diagram_pos, standard_colors[COLOR_BLACK]);
    diagram_pos.grow(-1, -1);

    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_WHITE);
    ctx.setTextAlign(0, 0);
    outText(ctx, diagram_pos.x, diagram_pos.getBottomY() - font_heights[FONT_SMALL], _("Warp \xE2\x86\x92"));
    diagram_pos.h -= font_heights[FONT_SMALL]+1;

    /* Diagram */
    int prev_ff = 0;
    int prev_x = 0;
    for (int warp = 1; warp <= 10; ++warp) {
        int new_x = warp*diagram_pos.w / 10;
        int new_ff = warp == 10 ? 1000 : getEngine(engine).getFuelFactor(warp) / (warp*warp);
        if (new_ff > 999)
            new_ff = 999;

        int delta_x = new_x - prev_x;

        /* draw the bars */
        for (int dx = 1; dx < delta_x; ++dx) {
            int this_ff    = (dx*new_ff + (delta_x-dx)*prev_ff) / delta_x;
            int this_color = COLOR_STATUS+15 - (16*this_ff)/1000;
            int this_y     = this_ff*diagram_pos.h/1000;
            ctx.setColor(this_color);
            drawVLine(ctx,
                      diagram_pos.x + dx + prev_x,
                      diagram_pos.getBottomY() - this_y,
                      diagram_pos.getBottomY());
        }
        prev_x = new_x;
        prev_ff = new_ff;
    }

    can.update(getExtent());
}

void
WEngineInfoDiagram::getLayoutInfo(LayoutInfo& info)
{
    /* The diagram is 2 (top) + 100 (size) + 1 (gap) + text + 2 (bottom)
       + 1 (gap) + 2*text */
    info.pref_size = info.min_size =
        GfxPoint(102, 106 + 3*font_heights[FONT_SMALL]);
    info.flags = 0;
}

void
WEngineInfoDiagram::setEngine(int n)
{
    if (n != engine) {
        engine = n;
        drawWidget(false);
    }
}

/*************************** WEngineStorageInfo **************************/

WEngineStorageInfo::WEngineStorageInfo()
    : UIWidget(), have(0), need(0)
{ }

void
WEngineStorageInfo::setEnginesAvailable(int n)
{
    if (n != have) {
        have = n;
        drawWidget(false);
    }
}

void
WEngineStorageInfo::setEnginesNeeded(int n)
{
    if (n != need) {
        need = n;
        drawWidget(false);
    }
}

void
WEngineStorageInfo::drawContent(GfxCanvas& can)
{
    int x = getExtent().x, y = getExtent().y, wi = getExtent().w;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setColor(UIColor::tc_White);
    ctx.setSolidBackground();
    outTextF(ctx, x, y, wi, format(_("Ship requires %d engine%!1{s%}."), need));
    y += font_heights[FONT_TITLE];

    ctx.useFont(*fonts[FONT_NORMAL]);
    outTextF(ctx, x, y, wi, format(_("In storage: %d engine%!1{s%}."), have));
    y += font_heights[FONT_NORMAL];

    can.update(getExtent());
}

void
WEngineStorageInfo::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size =
        font_sizes[FONT_NORMAL].scaledBy(15, 1).movedBy(0, font_heights[FONT_TITLE]);
    info.flags = 0;
}

/**************************** WShipBuildScreen ***************************/

/** \class WShipBuildScreen
    \brief Ship Build Screen, general version

    This class provides a complete ship build screen which allows the
    user to select a ship type. It does not provide buttons to confirm
    a build transaction. A derived class must add these. */

WShipBuildScreen::WShipBuildScreen(string_t name, GPlanet& the_planet)
    : UIWindow(name, UIWindow::BLUE_WINDOW, COLOR_DARK),
      hull_list(3000, 10, 15 * font_ems[FONT_NORMAL], &the_planet),
      engine_list(3001, 12 * font_ems[FONT_NORMAL], &the_planet),
      beam_list(3002, 12 * font_ems[FONT_NORMAL], &the_planet),
      torp_list(3003, 12 * font_ems[FONT_NORMAL], &the_planet),
      beam_count(false),
      torp_count(true),
      action(the_planet),
      item_cost_dpy(the_planet, false),
      cards(),
      tabs(cards),
      conn_update(the_planet.sig_changed.add(this, &WShipBuildScreen::onPlanetChanged))
{
    setLayout(UIVBoxLayout::instance5);

    add(h.add(new UIFocusIterator(1000, 1003, UIFocusIterator::fi_Horiz)));

    tabs.addTab(1000, UIKeyString(_("Starship Hulls")));
    tabs.addTab(1001, UIKeyString(_("Engines")));
    tabs.addTab(1002, UIKeyString(_("Beams")));
    tabs.addTab(1003, UIKeyString(_("Torpedoes")));

    /* Hulls */
    UIGroup& hull_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1000));
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i)
        if (int j = getTruehull(the_planet.getOwner(), i))
            hull_list.addHullId(j);
    hull_group.add(addFrame(h, hull_list));
    hull_group.add(hull_storage);
    hull_group.add(hull_info);
    hull_list.scrollToHull(action.getBuildOrder().hull);
    hull_list.sig_moved.add(this, &WShipBuildScreen::updateHullPage);

    /* Engines */
    UIGroup& engine_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1001));
    UIGroup& engine_vgroup = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& engine_hgroup = h.add(new UIGroup(UIHBoxLayout::instance5));
    engine_group.add(addFrame(h, engine_list));
    engine_group.add(engine_diag);
    engine_vgroup.add(engine_info);
    engine_vgroup.add(h.add(new UISpacer()));
    engine_vgroup.add(engine_storage);
    engine_vgroup.add(engine_hgroup);
    engine_hgroup.add(h.add(new UIButton(cm_BuildComponent, _("Spc-Build"), ' ')).addCommand());
    engine_hgroup.add(h.add(new UISpacer()));
    engine_group.add(engine_vgroup);
    engine_list.scrollTo(action.getBuildOrder().engine - 1, 1);
    engine_list.sig_moved.add(this, &WShipBuildScreen::updateEnginePage);

    /* Beams */
    UIGroup& beam_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1002));
    UIGroup& beam_vgroup = h.add(new UIGroup(UIVBoxLayout::instance0));
    beam_group.add(addFrame(h, beam_list));
    beam_group.add(beam_storage);
    beam_vgroup.add(beam_info);
    beam_vgroup.add(h.add(new UISpacer()));
    beam_vgroup.add(beam_count);
    beam_group.add(beam_vgroup);
    beam_list.scrollTo(action.getBuildOrder().beam_type - 1, 1);
    beam_list.sig_moved.add(this, &WShipBuildScreen::updateBeamPage);

    /* Torps */
    UIGroup& torp_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1003));
    UIGroup& torp_vgroup = h.add(new UIGroup(UIVBoxLayout::instance0));
    torp_group.add(addFrame(h, torp_list));
    torp_group.add(torp_storage);
    torp_vgroup.add(torp_info);
    torp_vgroup.add(h.add(new UISpacer()));
    torp_vgroup.add(torp_count);
    torp_group.add(torp_vgroup);
    torp_list.scrollTo(action.getBuildOrder().torp_type - 1, 1);
    torp_list.sig_moved.add(this, &WShipBuildScreen::updateLauncherPage);

    cards.add(hull_group);
    cards.add(engine_group);
    cards.add(beam_group);
    cards.add(torp_group);
    cards.sig_focus_changed.add(this, &WShipBuildScreen::updateBuildOrder);

    add(tabs);
    add(cards);

    /* Bill */
    UIGroup& bill_group = h.add(new UIGroup(UIHBoxLayout::instance5));
    bill_group.add(item_cost_dpy);
    bill_group.add(order_dpy);
    add(bill_group);

    hull_list.setFocusedWidget();
    updateHullPage();
    updateEnginePage();
    updateBeamPage();
    updateLauncherPage();
}

WShipBuildScreen::~WShipBuildScreen()
{ }

bool
WShipBuildScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_CountChanged) {
        updateBuildOrder();
        return true;
    } else if (cmd.code == cm_BuildComponent) {
        GPlanet& p = action.getPlanet();
        switch (tabs.getCurrentTab()) {
         case 1000:
            if (int slot = getTruehullSlot(p.getOwner(), hull_list.getCurrentHull()))
                doBuildComponent(getDisplayedTurn(), p.getId(),
                                 tl_Hull, slot,
                                 getHull(hull_list.getCurrentHull()));
            break;
         case 1001:
            doBuildComponent(getDisplayedTurn(), p.getId(),
                             tl_Engine, engine_list.getCurrent()+1,
                             getEngine(engine_list.getCurrent()+1));
            break;
         case 1002:
            doBuildComponent(getDisplayedTurn(), p.getId(),
                             tl_Beam, beam_list.getCurrent()+1,
                             getBeam(beam_list.getCurrent()+1));
            break;
         case 1003:
            doBuildComponent(getDisplayedTurn(), p.getId(),
                             tl_Torpedo, torp_list.getCurrent()+1,
                             getTorp(torp_list.getCurrent()+1));
            break;
        }
        return true;
    } else {
        return UIWindow::handleCommand(cmd, sender);
    }
}

bool
WShipBuildScreen::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code == 's' && tabs.getCurrentTab() == 1000) {
        /* 's' on hull tab: build ship query */
        GShipQuery q;
        q.setHullId(hull_list.getCurrentHull());
        q.setPlayerDisplaySet(GPlayerSet(action.getPlanet().getOwner()));
        q.setEngineId(engine_list.getCurrent()+1);
        q.setOwner(action.getPlanet().getOwner());
        q.complete(getDisplayedTurn().getCurrentUniverse());

        /* build dialog */
        WSpecView sv;
        sv.setQuery(q);
        sv.doStandardDialog();
        return true;
    } else {
        return UIWindow::handleEvent(event, second_pass);
    }
}

void
WShipBuildScreen::updateHullPage()
{
    const int hull = hull_list.getCurrentHull();
    const GPlanet& p = action.getPlanet();

    hull_info.setHullId(hull);
    if (int n = getTruehullSlot(p.getOwner(), hull))
        hull_storage.setCount(p.getBaseHullStoreSlot(n));
    else
        hull_storage.setCount(0);
    hull_storage.setImage(requestPixmap(ResId(res::ship, getHull(hull).getPicture(), hull)));

    engine_storage.setEnginesNeeded(getHull(hull).getNumEngines());
    beam_count.setPartsMax(getHull(hull).getMaxBeams());
    torp_count.setPartsMax(getHull(hull).getMaxLaunchers());

    updateBuildOrder();
}

void
WShipBuildScreen::updateBeamPage()
{
    const int beam = beam_list.getCurrent()+1;
    beam_info.setBeamId(beam);
    beam_storage.setCount(action.getPlanet().getBaseBeamStore(beam));
    beam_storage.setImage(requestPixmap(ResId("beam", beam)));
    updateBuildOrder();
}

void
WShipBuildScreen::updateLauncherPage()
{
    const int torp = torp_list.getCurrent()+1;
    torp_info.setTorpId(torp);
    torp_storage.setCount(action.getPlanet().getBaseLauncherStore(torp));
    torp_storage.setImage(requestPixmap(ResId("launcher", torp)));
    updateBuildOrder();
}

void
WShipBuildScreen::updateEnginePage()
{
    const int engine = engine_list.getCurrent()+1;
    const GPlanet& p = action.getPlanet();
    engine_info.setEngineId(engine);
    engine_diag.setEngine(engine);
    engine_storage.setEnginesAvailable(p.getBaseEngineStore(engine));
    updateBuildOrder();
}

void
WShipBuildScreen::updateBuildOrder()
{
    TBuildOrder bo;

    bo.hull   = hull_list.getCurrentHull();
    bo.engine = engine_list.getCurrent()+1;
    bo.beam_type  = beam_list.getCurrent()+1;
    bo.beam_count = beam_count.getPartsWanted();
    bo.torp_type  = torp_list.getCurrent()+1;
    bo.torp_count = torp_count.getPartsWanted();
    bo.unused = 0;
    action.setBuildOrder(bo);
    order_dpy.setBuildOrder(bo);
    item_cost_dpy.setTotalCost(action.getCost());

    // since this might be called during destruction, we cannot use
    // 'tabs.getCurrentTab()'. 'tabs' might already be gone.
    if (UIBaseWidget* p = cards.getFocusedChild()) {
        switch (p->getId()) {
         case 1000: {
            item_cost_dpy.setPartCost(getHull(hull_list.getCurrentHull()).getCost());
            item_cost_dpy.setPartTech(action.getPlanet().getBaseTech(tl_Hull),
                                      getHull(hull_list.getCurrentHull()).getTech());
            break;
         }
         case 1001: {
            GEngine& e = getEngine(engine_list.getCurrent()+1);
            item_cost_dpy.setPartCost(e.getCost());
            item_cost_dpy.setPartTech(action.getPlanet().getBaseTech(tl_Engine), e.getTech());
            break;
         }
         case 1002: {
            GBeam& b = getBeam(beam_list.getCurrent()+1);
            item_cost_dpy.setPartCost(b.getCost());
            item_cost_dpy.setPartTech(action.getPlanet().getBaseTech(tl_Beam), b.getTech());
            break;
         }
         case 1003: {
            GTorpLauncher& t = getTorp(torp_list.getCurrent()+1);
            item_cost_dpy.setPartCost(t.getCost());
            item_cost_dpy.setPartTech(action.getPlanet().getBaseTech(tl_Torpedo), t.getTech());
            break;
         }
        }
    }
}

void
WShipBuildScreen::onPlanetChanged(int)
{
    item_cost_dpy.setTotalCost(action.getCost());
    if (UIBaseWidget* p = cards.getFocusedChild()) {
        switch (p->getId()) {
         case 1000:
            updateHullPage();
            hull_list.drawWidget(false);
            break;
         case 1001:
            updateEnginePage();
            engine_list.drawWidget(false);
            break;
         case 1002:
            updateBeamPage();
            beam_list.drawWidget(false);
            break;
         case 1003:
            updateLauncherPage();
            torp_list.drawWidget(false);
            break;
        }
    }
}

/*********************** WShipBuildScreenWithActor ***********************/

WShipBuildScreenWithActor::WShipBuildScreenWithActor(GPlanet& the_planet)
    : WShipBuildScreen(_("Build a Ship"), the_planet),
      the_planet(the_planet)
{
    // UIGroup[UIHBoxLayout]
    //   UIButton "U"
    //   UIStaticText "Use parts from storage"
    // UIGroup[UIHBoxLayout]
    //   UIButton "Enter"
    //   UIButton "ESC"
    //   [UIButton "Cancel"]
    //   UISpacer
    //   [UIButton "Help"]
    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));

    UIButton& ok = h.add(new UIButton(2000, _("Enter - Build"), SDLK_RETURN));
    UIButton& cancel = h.add(new UIButton(2001, _("ESC - Exit"), SDLK_ESCAPE));
    g.add(ok);
    g.add(cancel);
    if (the_planet.getBaseBuildHull()) {
        UIButton& cancel_build = h.add(new UIButton(2002, UIKeyString(_("C - Cancel Build"))));
        cancel_build.sig_fired.add(this, &WShipBuildScreenWithActor::onCancelBuild);
        g.add(cancel_build);
    }
    g.add(h.add(new UISpacer()));

    UIButton& u = h.add(new UIButton(2003, "U", 'u'));
    g2.add(u);
    g2.add(h.add(new UIStaticText(_("Use parts from storage"), UIStaticText::WhiteTitle)));
    g2.add(h.add(new UISpacer()));

    add(g2);
    add(g);

    ok.sig_fired.add(this, &WShipBuildScreenWithActor::onOk);
    cancel.sig_fired.add(this, &WShipBuildScreenWithActor::onCancel);
    u.sig_fired.add(this, &WShipBuildScreenWithActor::onToggleUseParts);

    updateUseParts();
    pack();
}

WShipBuildScreenWithActor::~WShipBuildScreenWithActor()
{ }

void
WShipBuildScreenWithActor::onOk()
{
    GPlanetTransfer transfer(the_planet);
    GStarbaseBuildTransaction ta(action, transfer);
    if (GError e = ta.getError()) {
        messageBox(format(_("You can not build this ship: %s"), e.getUserError()),
                   _("Build order rejected"));
        ta.doCancel();
        transfer.doCancel();
    } else {
        ta.doCommit();
        transfer.doCommit();
        action.getPlanet().setBaseBuildOrder(action.getBuildOrder());
        stop(1);
    }
}

void
WShipBuildScreenWithActor::onCancel()
{
    stop(0);
}

void
WShipBuildScreenWithActor::onCancelBuild()
{
    static const TBuildOrder blank = { 0, 0, 0, 0, 0, 0, 0 };
    action.getPlanet().setBaseBuildOrder(blank);
    stop(1);
}

void
WShipBuildScreenWithActor::onToggleUseParts()
{
    action.setUsePartsFromStorage(!action.isUsePartsFromStorage());
    updateUseParts();
    onPlanetChanged(0); // FIXME: should use signals properly.
}

void
WShipBuildScreenWithActor::updateUseParts()
{
    setStateOf(2003, UIButton::bf_Highlight, action.isUsePartsFromStorage());
}
