/**
  *  \file game/shipquery.cc
  *  \brief Ship Query
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/shipquery.h"
#include "game/univ.h"
#include "game/team.h"
#include "game/ship.h"
#include "game/config.h"
#include "game/hullfunc.h"


/*
  Item                  Derived from
  --------------------  --------------------
  hull_id               ship_id
  ship_id               -
  filter_level_set      defaults to all players
  display_level_set     -
  filter_player_set     defaults to all players
  display_player_set    owner
  engine_id             ship_id
  combat_mass           ship_id or hull_id, optionally engine_id
  crew                  ship_id or hull_id
  owner                 ship_id or display_player_set, fallback to global player Id

  That is, to display information about a single ship, one has to fill in just
  ship_id, display_level_set, owner or display_player_set; the rest is derived by complete().
 */

/** Constructor. Makes a blank object. */
GShipQuery::GShipQuery()
{
    clear();
}

/** Clear. Makes this object blank. */
void
GShipQuery::clear()
{
    hull_id            = 0;
    ship_id            = 0;
    filter_level_set   = GExpLevelSet::fromInteger(-1U);
    display_level_set  = GExpLevelSet::fromInteger(1);
    filter_player_set  = GPlayerSet::fromInteger(-1U);
    display_player_set = GPlayerSet::fromInteger(0);
    engine_id          = 0;
    combat_mass        = 0;
    crew               = 0;
    owner              = 0;
}

/** Initialize for existing ship.
    \param univ [in] Universe
    \param sid  [in] Ship Id */
void
GShipQuery::initForExistingShip(const GUniverse& univ, int sid)
{
    clear();

    ship_id = sid;

    owner = univ.getShip(sid).getOwner();

    int16_t level = univ.getShip(ship_id).unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
    if (level >= 0)
        display_level_set = GExpLevelSet(level);

    complete(univ);
}

/** Complete request. If this request is (partially) filled, it will
    be completed as far as possible. */
void
GShipQuery::complete(const GUniverse& univ)
{
    if (univ.isValidShipId(ship_id) && univ.getShip(ship_id).isExistant()) {
        /* We know the ship Id, so try to complete the request from the real universe */
        const GShip& sh = univ.getShip(ship_id);

        if (hull_id <= 0 && sh.getHullId() > 0)
            hull_id = sh.getHullId();

        if (engine_id <= 0 && sh.getEngine().isKnown())
            engine_id = sh.getEngine();

        if (crew <= 0 && sh.getCrew().isKnown())
            crew = sh.getCrew();

        if (owner <= 0 && sh.getRealOwner() > 0)
            owner = sh.getRealOwner();
    }

    /* Do we have an owner? */
    if (owner <= 0) {
        /* Try to derive from display_player_set. FIXME: needed? */
        for (int i = 1; i <= NUM_OWNERS; ++i)
            if (display_player_set == GPlayerSet(i))
                owner = i;
    }
    if (owner <= 0) {
        /* Use global default */
        owner = getPlayerId();
    }

    /* If we have an owner, try to complete display_player_set */
    if (owner > 0 && display_player_set.empty()) {
        display_player_set = GPlayerSet(owner);
    }

    /* Crew */
    if (crew <= 0 && hull_id > 0) {
        crew = getHull(hull_id).getCrew();
    }

    /* Now figure out combat mass */
    if (combat_mass <= 0 && hull_id > 0) {
        combat_mass = getHull(hull_id).getMass();

        /* Fed bonus */
        if (raceId(owner) == 1 && config.AllowFedCombatBonus())
            combat_mass += 50;

        /* ESB */
        int32_t esb = 0;
        if (config.AllowEngineShieldBonus())
            esb += config.EngineShieldBonusRate(owner);

        if (univ.isValidShipId(ship_id) && univ.getShip(ship_id).isExistant())
            esb += getExperienceBonus(config.EModEngineShieldBonusRate,
                                      univ.getShip(ship_id).unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel)));

        if (engine_id > 0)
            combat_mass += getEngine(engine_id).getCost().get(el_Money) * esb / 100;
    }
}

/** Enumerate ship specials for this ship. */
void
GShipQuery::enumShipSpecials(const GUniverse& univ, GHullFunctionList& list, bool includeRacialAbilities) const
{
    // hullfunc.pas::EnumHullfuncs
    if (hull_id == 0 || hull_id > getNumHulls()) {
        /* Invalid hull selection, blank result */
    } else if (univ.isValidShipId(ship_id)) {
        /* We have a ship */
        ::getHull(hull_id).enumerateHullFunctions(list, filter_player_set, filter_level_set, false, includeRacialAbilities);
        univ.getShip(ship_id).enumShipSpecificSpecials(list);
    } else {
        /* We don't have a ship, so list defaults */
        ::getHull(hull_id).enumerateHullFunctions(list, filter_player_set, filter_level_set, true, includeRacialAbilities);
    }
}

/** Compare for equality. */
bool
GShipQuery::operator==(const GShipQuery& other) const
{
    return hull_id            == other.hull_id
        && ship_id            == other.ship_id
        && filter_level_set   == other.filter_level_set
        && display_level_set  == other.display_level_set
        && filter_player_set  == other.filter_player_set
        && display_player_set == other.display_player_set
        && engine_id          == other.engine_id
        && combat_mass        == other.combat_mass
        && crew               == other.crew
        && owner              == other.owner;
}
