/**
  *  \file int/if/mineif.cc
  *  \brief Interpreter Game Interface: Minefields
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/mineif.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/mine.h"
#include "int/compound.h"
#include "int/data.h"
#include "int/if/ifutil.h"
#include "int/if/playerif.h"
#include "int/interror.h"
#include "int/value.h"
#include "int/args.h"
#include "cpluslib/format.h"
#include "int/contextio.h"

/** Implementation of "Minefield(nn)".
    \param args function parameters
    \return value representing "Minefield(nn)" */
IntValue*
IFMinefieldGet(IntArgBlock& args)
{
    int32_t mid;

    args.checkArgumentCount(1);
    if (!checkIntArg(mid, args.getNext()))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return new IntMinefieldContext(mid);
}

/** Implementation of "ForEach Minefield".
    \return Context representing first minefield */
IntContext*
IFMinefieldMake()
{
    if (!haveDisplayedTurn())
        return 0;

    int mid = global_minefield_type.getNextNoWrap(0, false);
    if (mid != 0)
        return new IntMinefieldContext(mid);
    else
        return 0;
}

/************************** IntMinefieldContext **************************/

namespace {
    enum { MinefieldProperty, OwnerProperty };
}

static const TIntMapping minefield_mapping[] = {
    { "ID",         impId,        MinefieldProperty },
    { "LASTSCAN",   impLastScan,  MinefieldProperty },
    { "LOC.X",      impLocX,      MinefieldProperty },
    { "LOC.Y",      impLocY,      MinefieldProperty },
    { "OWNER",      iplShortName, OwnerProperty },
    { "OWNER$",     iplId,        OwnerProperty },
    { "OWNER.ADJ",  iplAdjName,   OwnerProperty },
    { "RADIUS",     impRadius,    MinefieldProperty },
    { "SCANNED",    impScanType,  MinefieldProperty },
    { "TYPE",       impTypeStr,   MinefieldProperty },
    { "TYPE$",      impTypeCode,  MinefieldProperty },
    { "UNITS",      impUnits,     MinefieldProperty },
};

/** Constructor.
    \param mid Minefield Id */
IntMinefieldContext::IntMinefieldContext(int mid)
    : mid(mid)
{ }

/** Destructor. */
IntMinefieldContext::~IntMinefieldContext()
{ }

int32_t
IntMinefieldContext::lookup(string_t name)
{
    return lookupName(name, minefield_mapping, countof(minefield_mapping));
}

void
IntMinefieldContext::set(int32_t index, IntValue* value)
{
    if (!global_minefield_type.isExistant(mid))
        throw IntError::notAssignable();
    switch (minefield_mapping[index].domain) {
     case MinefieldProperty:
        setMinefieldProperty(global_minefield_type.getObject(mid), IntMinefieldProperty(minefield_mapping[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntMinefieldContext::get(int32_t index)
{
    if (!global_minefield_type.isExistant(mid))
        return 0;

    const GMinefield& mf = global_minefield_type.getObject(mid);
    switch (minefield_mapping[index].domain) {
     case MinefieldProperty:
        return getMinefieldProperty(mf, IntMinefieldProperty(minefield_mapping[index].index));
     case OwnerProperty:
        return getPlayerProperty(mf.getOwner(), IntPlayerProperty(minefield_mapping[index].index));
    }
    return 0;
}

bool
IntMinefieldContext::next()
{
    int n = global_minefield_type.getNextNoWrap(mid, false);
    if (n != 0) {
        mid = n;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntMinefieldContext::clone()
{
    return new IntMinefieldContext(mid);
}

GObject*
IntMinefieldContext::getObject()
{
    if (global_minefield_type.isExistant(mid))
        return &global_minefield_type.getObject(mid);
    else
        return 0;
}

string_t
IntMinefieldContext::toString(bool /*readable*/)
{
    return format("Minefield(%d)", mid);
}

void
IntMinefieldContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Minefield;
    sv.value = mid;
}

/** Get minefield property.
    \param mf Minefield to query
    \param imp Property to query
    \return property value, newly allocated */
IntValue*
getMinefieldProperty(const GMinefield& mf, IntMinefieldProperty imp)
{
    switch (imp) {
     case impId:
        return makeIntValue(mf.getId());
     case impLastScan:
        return makeIntValue(mf.getTurnLastSeen());
     case impLocX:
        return makeIntValue(mf.getPos().x);
     case impLocY:
        return makeIntValue(mf.getPos().y);
     case impRadius:
        return makeIntValue(mf.getRadius());
     case impScanType:
        // FIXME: PCC 1.x checks for <0. Can this happen?
        return makeIntValue(mf.getAction());
     case impTypeCode:
        return makeBoolValue(mf.isWeb());
     case impTypeStr:
        return makeStringValue(mf.isWeb() ? "Web Mines" : "Mines");
     case impUnits:
        return makeIntValue(mf.getUnits());
    }
    return 0;
}

/** Set minefield property. */
void
setMinefieldProperty(GMinefield& /*mf*/, IntMinefieldProperty /*imp*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}
