/**
  *  \file tools/c2dumpspec.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "tools/c2dump.h"
#include "cpluslib/format.h"
#include "cpluslib/buffer.h"
#include "io/streams.h"

void
parseHullSpec(Stream& s, DumpOutputReceiver& r)
{
    // Read file into memory
    Buffer<char> buf;
    s.readBuffer(buf);
    DumpBlockInput bi(buf.ptr(), buf.getSize());

    // All bytes used, so parse all hulls we find
    int n = 0;
    while (bi.getRemainingSize() >= 60) {
        ++n;
        r.startRecord(format("Hull(%d)", n));
        r.addField("Name", bi.readString(30));
        r.addField("Picture", bi.readWord());
        r.addField("Unused", bi.readWord());
        r.addField("Tritanium", bi.readWord());
        r.addField("Duranium", bi.readWord());
        r.addField("Molybdenum", bi.readWord());
        r.addField("Fuel_capacity", bi.readWord());
        r.addField("Crew", bi.readWord());
        r.addField("Num_engines", bi.readWord());
        r.addField("Mass", bi.readWord());
        r.addField("Tech_level", bi.readWord());
        r.addField("Cargo_capacity", bi.readWord());
        r.addField("Num_bays", bi.readWord());
        r.addField("Max_tubes", bi.readWord());
        r.addField("Max_beams", bi.readWord());
        r.addField("Cost", bi.readWord());
        r.endRecord();
    }

    // Remainder
    dumpRemainder(bi, r);
}

void
parseTorpSpec(Stream& s, DumpOutputReceiver& r)
{
    // Read file into memory
    Buffer<char> buf;
    s.readBuffer(buf);
    DumpBlockInput bi(buf.ptr(), buf.getSize());

    // torpspec contains 10x38 bytes, but is usually larger
    for (int slot = 1; slot <= 10 && bi.getRemainingSize() > 0; ++slot) {
        r.startRecord(format("Tube(%d)", slot));
        r.addField("Name", bi.readString(20));
        r.addField("Torp_cost", bi.readWord());
        r.addField("Cost", bi.readWord());
        r.addField("Tritanium", bi.readWord());
        r.addField("Duranium", bi.readWord());
        r.addField("Molybdenum", bi.readWord());
        r.addField("Mass", bi.readWord());
        r.addField("Tech_level", bi.readWord());
        r.addField("Kill_power", bi.readWord());
        r.addField("Damage_power", bi.readWord());
        r.endRecord();
    }

    dumpRemainder(bi, r);

}

void
parseBeamSpec(Stream& s, DumpOutputReceiver& r)
{
    // Read file into memory
    Buffer<char> buf;
    s.readBuffer(buf);
    DumpBlockInput bi(buf.ptr(), buf.getSize());

    // beamspec contains 10x36 bytes
    for (int slot = 1; slot <= 10 && bi.getRemainingSize() > 0; ++slot) {
        r.startRecord(format("Beam(%d)", slot));
        r.addField("Name", bi.readString(20));
        r.addField("Cost", bi.readWord());
        r.addField("Tritanium", bi.readWord());
        r.addField("Duranium", bi.readWord());
        r.addField("Molybdenum", bi.readWord());
        r.addField("Mass", bi.readWord());
        r.addField("Tech_level", bi.readWord());
        r.addField("Kill_power", bi.readWord());
        r.addField("Damage_power", bi.readWord());
        r.endRecord();
    }

    dumpRemainder(bi, r);
}

void
parseEngSpec(Stream& s, DumpOutputReceiver& r)
{
    // Read file into memory
    Buffer<char> buf;
    s.readBuffer(buf);
    DumpBlockInput bi(buf.ptr(), buf.getSize());

    // engspec contains 9x66 bytes, but is usually larger
    for (int slot = 1; slot <= 10 && bi.getRemainingSize() > 0; ++slot) {
        r.startRecord(format("Engine(%d)", slot));
        r.addField("Name", bi.readString(20));
        r.addField("Cost", bi.readWord());
        r.addField("Tritanium", bi.readWord());
        r.addField("Duranium", bi.readWord());
        r.addField("Molybdenum", bi.readWord());
        r.addField("Tech", bi.readWord());
        for (int i = 1; i <= 9; ++i)
            r.addField(format("Fuel_factor(%d)", i), bi.readLong());
        r.endRecord();
    }

    dumpRemainder(bi, r);
}

void
parseTrueHull(Stream& s, DumpOutputReceiver& r)
{
    // Read file into memory
    Buffer<char> buf;
    s.readBuffer(buf);
    DumpBlockInput bi(buf.ptr(), buf.getSize());

    // truehull contains 11x20 words, but is sometimes much larger
    for (int pl = 1; pl <= 11 && bi.getRemainingSize() > 0; ++pl) {
        r.startRecord(format("Player(%d)", pl));
        for (int slot = 1; slot <= 20; ++slot)
            r.addField(format("Slot(%d)", slot), bi.readWord());
        r.endRecord();
    }

    dumpRemainder(bi, r);
}

void
parseNameList(Stream& s, DumpOutputReceiver& r)
{
    // Read file into memory
    Buffer<char> buf;
    s.readBuffer(buf);
    DumpBlockInput bi(buf.ptr(), buf.getSize());

    r.startRecord("Names");
    int n = 0;
    while (bi.getRemainingSize() >= 20) {
        ++n;
        r.addField(format("Name(%d)", n), bi.readString(20));
    }
    r.endRecord();

    dumpRemainder(bi, r);
}

void
parseXYPlan(Stream& s, DumpOutputReceiver& r)
{
    // Read file into memory
    Buffer<char> buf;
    s.readBuffer(buf);
    DumpBlockInput bi(buf.ptr(), buf.getSize());

    int n = 0;
    while (bi.getRemainingSize() >= 6) {
        ++n;
        r.startRecord(format("Planet(%d)", n));
        r.addField("Location", bi.readCoordinate());
        r.addField("Owner", bi.readWord());
        r.endRecord();
    }

    dumpRemainder(bi, r);
}

void
parseRaceNames(Stream& s, DumpOutputReceiver& r)
{
    // Read file into memory
    Buffer<char> buf;
    s.readBuffer(buf);
    DumpBlockInput bi(buf.ptr(), buf.getSize());

    r.startRecord("Long_names");
    for (int i = 1; i <= 11; ++i)
        r.addField(format("Player(%d)", i), bi.readString(30));
    r.endRecord();

    r.startRecord("Short_names");
    for (int i = 1; i <= 11; ++i)
        r.addField(format("Player(%d)", i), bi.readString(20));
    r.endRecord();

    r.startRecord("Adjectives");
    for (int i = 1; i <= 11; ++i)
        r.addField(format("Player(%d)", i), bi.readString(12));
    r.endRecord();

    dumpRemainder(bi, r);
}

